#!/usr/bin/perl
###############################################################################
# SIS File Dumper
# -----------------------------------------
# Copyright 2003-2004 Jimmy Shah  All rights reserved.
###############################################################################
# This program is licensed under the GNU Public License, version 2 or later.
# A copy of this license is included in the package as License.txt.
# If it is missing, a copy is available at http://www.gnu.org/copyleft/gpl.html
###############################################################################
# Bugfixes
# - 2004-06-15
#   - fixed bang path 
#   - fixed case of File::Spec
#   - fixed uncompress always; -e redundant test
# reported by Patrick R.Mullen(Mullen Scientific Software) & Joe Stewart(LUHRQ)
###############################################################################
# - 2004-06-19
#   - fix values for conditional variables
#     holes filled with "InvalidConditionVariable"
###############################################################################
# TODO
# - Java related extensions
#
###############################################################################


use 5.6.0;
use strict;
use warnings;
use File::Basename;
use File::Spec;
use Getopt::Std;
use Disassemble::X86::MemRegion;
use Compress::Zlib;
use Unicode::String qw(latin1);
use diagnostics;
	our ($opt_e,$opt_d,$opt_o,$opt_s,$opt_t,$opt_z);
	our $basepath;
	my $inpfile;
	my $SISbyteStream; 

getopts('eotszd:');
&getbasepath;
if (@ARGV == 0) {&usage};
$inpfile = $ARGV[0];
#############################################################
# Pop whole file into memory
#############################################################
{
my $inputfile;
open($inputfile,"<$inpfile") or die "Cannot open $inpfile \n\n";
	binmode $inputfile;	
		{
		local $/;			#undef input record seperator
		$SISbyteStream = <$inputfile>; #gives one big record
		}
close $inputfile;
unless ($SISbyteStream) {die "Error reading ($inputfile)\n"};
}
#############################################################
# use Disassemble::X86::MemRegion ; startup is slow but beats
# redoing the work
#############################################################
	my $E32file = Disassemble::X86::MemRegion->new( mem => $SISbyteStream );
    my %SISDump; # specific offset hash
    my @SISLanguages; # language array
    my @SISFfiles;
	my ($firstFIle,$DmpFileCount);
    FillDump(\%SISDump,$E32file,\@SISLanguages,\@SISFfiles);			# fill the dump
	#@SISFfiles = reverse @SISFfiles; 
	shift @SISFfiles; # remove empty entry after reversal
	Language2Txt(\%SISDump,\@SISLanguages);               #hext to text
    Type2Txt(\%SISDump);
    PrintHeader(\%SISDump);                              #dump header
    PrintListFiles($SISDump{NumFiles},\@SISFfiles);   #print destination names
    if ($opt_d && $SISDump{NotStub}) {              # handle file dump args
	$opt_d =~ m/(\d+)-?(\d*)/;                                   ##################################
	DumpSomeFiles($E32file,\%SISDump,$1,($2?$2-$1:0),\@SISFfiles);  # the real purpose of this tool
	}                                                            ##################################  



####################################################################
# Subs
####################################################################

sub usage {
		die <<END_USAGE
DumpSIS - Symbian SIS File Dumper ver. 0.81
Copyright 2003-2004 Jimmy Shah  All rights reserved.

Usage: $0 [-oetsd #] filename

Options:

    -d  # is number of file in package to dump to disk
        A range can also be specified. (e.g. -d 2-12)
    -o  Overwrite existing file with dumped file
    -e  Display extended information
    -t  Display file type information
    -s  Display source filenames

    No options dumps only header and file list.

END_USAGE
}

sub getbasepath
{
my ($file,$suffix);
($file,$basepath,$suffix) = fileparse($0,("\.pl"));
}

sub FillDump
{
#########################################
# fill the dump
#########################################
     my ($Dump,$File,$Languages,$PkgFiles) = @_;
     $Dump->{Uid1} =        			$File->get_long(0x0);
     $Dump->{Uid2} =        			$File->get_long(0x4);
     unless ($Dump->{Uid2} == 0x1000006D || $Dump->{Uid2} ==0x10003A12) 
     {print "\n*****Not valid Symbian SIS file.*****\n\n";&usage}
     $Dump->{Uid3} =        			$File->get_long(0x8);
     $Dump->{Checksum} =       	  	 	$File->get_long(0xc);
     $Dump->{CRCPackage} =         		$File->get_word(0x10);
     $Dump->{NumLanguages}  =      		$File->get_word(0x12);
     $Dump->{NumFiles} =    			$File->get_word(0x14);
     $Dump->{NumDependencies} =	   		$File->get_word(0x16);
     $Dump->{InstallLanguage}	 = 		$File->get_word(0x18);
     $Dump->{InstalledFiles} 	=  		$File->get_word(0x1a);
     $Dump->{InstallDrive} =   	   		$File->get_word(0x1c);
     $Dump->{NumCapabilities} =			$File->get_word(0x1e);
     $Dump->{InstallerVersion} =   		$File->get_long(0x20);
     $Dump->{Options} =    				$File->get_word(0x24);
     $Dump->{Type} =    		   		$File->get_word(0x26);
     $Dump->{MajorVersion} =    		$File->get_word(0x28);
     $Dump->{MinorVersion} =    		$File->get_word(0x2a);
     $Dump->{Variant} =    				$File->get_long(0x2c);
     $Dump->{LanguagesOffset} =			$File->get_long(0x30);
     $Dump->{FilesOffset} =    	   		$File->get_long(0x34);
     $Dump->{DependenciesOffset} = 		$File->get_long(0x38);
     $Dump->{ComponentNameOffset} = 	$File->get_long(0x40);
     $Dump->{NotStub} = 1;
     if ($Dump->{Uid2} == 0x10003A12)
     {
     # per PSP, the unicode build for ER5 never made it out the door
     print "Unicode Build\n";
     $Dump->{Unicode} = 1;
     $Dump->{ComponentNameOffset2} = 	$File->get_long(0x48);
     $Dump->{OriginalFileLength} = 		$File->get_long(0x50);
     }
     #Stub File has installation language filled in
     if ($Dump->{InstallLanguage}){print "Stub SIS File\n";$Dump->{NotStub} = 0}
     #Stub file for ROM Instalation
     if (lc(chr $Dump->{InstallDrive}) eq "z" ){print "ROM Stub SIS File\n";$Dump->{NotStub} = 0}

     # Language look up array
     $$Languages[0x0] = "Test";$$Languages[0x1] = "UK English";$$Languages[0x2] = "French";
     $$Languages[0x3] = "German";$$Languages[0x4] = "Spanish";$$Languages[0x5] = "Italian";
     $$Languages[0x6] = "Swedish";$$Languages[0x7] = "Danish";$$Languages[0x8] = "Norwegian";
     $$Languages[0x9] = "Finnish";$$Languages[0xA] = "American English";
     $$Languages[0xB] = "Swiss French";$$Languages[0xC] = "Swiss German";
     $$Languages[0xD] = "Portuguese";$$Languages[0xE] = "Turkish";
     $$Languages[0xF] = "Icelandic";$$Languages[0x10] = "Russian";
     $$Languages[0x11] = "Hungarian";$$Languages[0x12] = "Dutch";
     $$Languages[0x13] = "Belgian Flemish";$$Languages[0x14] = "Australian English";
     $$Languages[0x15] = "Belgian French";$$Languages[0x16] = "Austrian German";
     $$Languages[0x17] = "New Zealand English";$$Languages[0x18] = "International French";
     $$Languages[0x19] = "Czech";$$Languages[0x1A] = "Slovak";$$Languages[0x1B] = "Polish";
     $$Languages[0x1C] = "Slovenian";$$Languages[0x1D] = "Taiwan Chinese";
     $$Languages[0x1E] = "Hong Kong Chinese";$$Languages[0x1F] = "PRC Chinese";
     $$Languages[0x20] = "Japanese";$$Languages[0x21] = "Thai";

     # Install files array
     my ($m,$FileRecords,$offset,@P);
     # $FileRecords = file remaining count, $offset = current offset
     for ($FileRecords = $Dump->{NumFiles},$offset = $Dump->{FilesOffset};$FileRecords>=1;$FileRecords--)
     {
     	@P = GetPkgFileLine($Dump,$File,$offset);
     	$offset = pop @P;
     	#fill array in reverse order
     	$$PkgFiles[$FileRecords] = [@P];
     }
               
}

sub PrintHeader
{
my $Dump = shift;
printf "Uid1: %#8x Uid2: %#8x Uid3: %#8x  Uid4: %#8x\nSIS CRC: %#4x\n",
$Dump->{Uid1},$Dump->{Uid2},$Dump->{Uid3},$Dump->{Checksum},$Dump->{CRCPackage};
printf "Number of Languages: %d\n",$Dump->{NumLanguages};
if($opt_e) {PrintSupportedLanguages($Dump,$E32file,\@SISLanguages)}
printf "Number of Files: %s\n", $Dump->{NumFiles};
printf "Number of Dependencies: %d\n", $Dump->{NumDependencies};
printf "Installed Language: %s\n",$Dump->{InstallLanguage};
printf "Last Installed File: %d\nInstalled Drive: %s\n",
$Dump->{InstalledFiles},chr $Dump->{InstallDrive};
printf "Installer Version: %d (%#2x)\n\n",$Dump->{InstallerVersion},$Dump->{InstallerVersion};
PrintOptions($Dump);
printf "Type: %s\t",$Dump->{Type};
printf "Version: %d.%d.%d\n",$Dump->{MajorVersion},$Dump->{MinorVersion},$Dump->{Variant};
if($opt_e) 
{
if ($Dump->{NumDependencies}){PrintDependencies($Dump,$E32file)}
}
if($opt_e) {PrintComponentNames($Dump,$E32file)}
}

sub Language2Txt
{
my ($Dump,$Languages) = @_;
   if ($Dump->{InstallLanguage})
   {
   $Dump->{InstallLanguage} = $$Languages[$Dump->{InstallLanguage}];
   }
}

sub PrintSupportedLanguages
{
my ($Dump,$File,$Languages) = @_;
my ($n,$offset);
   for ($n = $Dump->{NumLanguages},$offset = $Dump->{LanguagesOffset};$n>=1;$n--,$offset+=2)
   {
    printf "\t\|\-%s\n",$$Languages[$File->get_word($offset)];
   }
}


sub PrintDependencies
{
###########################################################
#  Purpose of this tool is not to duplicate install program
#  This info is not stored.
###########################################################
my ($Dump,$File) = @_;
my ($UID,$Major,$Minor,$variant,$offset,$n,$records);
   print "\nDependencies\n------------\n";
   for ($offset = $Dump->{DependenciesOffset},$n = 0;$n < $Dump->{NumDependencies};$n++)
   {
    #################
    #  get UID
    #################
       $UID = $File->get_long($offset); $offset += 4;
    #################
    #  get version Major
    #################
       $Major = $File->get_word($offset); $offset += 2;
    #################
    #  get version Minor
    #################
       $Minor = $File->get_word($offset); $offset += 2;
    #################
    #  get variant
    #################
       $variant = $File->get_long($offset); $offset += 4;
    #################
    #  dependencies
    #################
      my (@nameLengths,@names,$f);
      for ($records=$Dump->{NumLanguages},$f=1;$records > 0;$records--,$f++)
      {
         $nameLengths[$f]= $File->get_long($offset); $offset += 4;
      }

      for ($records=$Dump->{NumLanguages},$f=1;$records > 0;$records--,$f++)
      {

		$names[$f]= get_string_got_lenlong($File,$File->get_long($offset),$nameLengths[$f]); $offset += 4;
		if ($Dump->{Unicode})
		{
		      $names[$f] = Unicode::String->new($names[$f]);
		      $names[$f] = latin1($names[$f]);
		      $names[$f] =~ s/\c@//g;
		}
      }


    #############################

    printf "%d) UID: %#8x\n   Version: %d.%0d\n   Variant: %#x\n",$n+1,$UID,$Major,$Minor,$variant;
    for($f=1;$f<=$Dump->{NumLanguages};$f++)
    {
    printf "   |-%s\n",$names[$f];
    }
    print "\n";
   }

}

sub PrintComponentNames
{
my ($Dump,$File) = @_;
my (@nameLengths,@names,$offset,$records,$f);

#################
#  componentName
#################
  for ($offset = $Dump->{ComponentNameOffset},$records=$Dump->{NumLanguages},$f=1;$records > 0;$records--,$f++)
  {
     $nameLengths[$f]= $File->get_long($offset); $offset += 4;
  }

  for ($records=$Dump->{NumLanguages},$f=1;$records > 0;$records--,$f++)
  {

      $names[$f] = get_string_got_lenlong($File,$File->get_long($offset),$nameLengths[$f]); $offset += 4;
      if ($Dump->{Unicode})
      {
	      $names[$f] = Unicode::String->new($names[$f]);
	      $names[$f] = latin1($names[$f]);
	      $names[$f] =~ s/\c@//g;
      }
  }
    print "Component Name\n--------------\n";
    for($f=1;$f<=$Dump->{NumLanguages};$f++)
    {
    printf "|-%s\n",$names[$f];
    }
}


sub GetPkgFileLine
{
# file offset $offset
my ($Dump,$File,$offset,$pkglinetype) = @_;
my ($filetype,$filedetails,$Sourcename,$Destname,$counter1,@filesizes,@fileoffsets,@filecompoffsets);

		# flags - fileflag2text useless
		$pkglinetype = $File->get_long($offset);$offset += 4;

    if ($pkglinetype <= 1) # standard files or language dependent files
    {
	    #Type and details
		($filetype,$filedetails) = PkgLine2Txt($File->get_long($offset),$File->get_long($offset+4));
		$offset+=8;
		$Sourcename = get_string_got_lenlong($File,$File->get_long($offset+4),$File->get_long($offset));
		$offset+=8;
			      if ($Dump->{Unicode})
			      {
			      $Sourcename = Unicode::String->new($Sourcename);
			      $Sourcename = latin1($Sourcename);
			      $Sourcename =~ s/\c@//g;
		           }
		$Destname = get_string_got_lenlong($File,$File->get_long($offset+4),$File->get_long($offset));
		$offset+=8;
			      if ($Dump->{Unicode})
			      {
			      $Destname = Unicode::String->new($Destname);
			      $Destname = latin1($Destname);
			      $Destname =~ s/\c@//g;
		           }
		
		for ($counter1 = $Dump->{NumLanguages};$counter1 >=1;$counter1--)
		{
			if ($pkglinetype)
			{
				$filesizes[$counter1] = $File->get_long($offset); $offset += 4;
			}
			else
			{
				$filesizes[1] = $File->get_long($offset); $offset += 4;$counter1 = 0;
			}
		}
		
		for ($counter1 = $Dump->{NumLanguages};$counter1 >=1;$counter1--)
		{
			if ($pkglinetype)
			{
				$filecompoffsets[$counter1] = $File->get_long($offset); $offset += 4;
			}
			else
			{
				$filecompoffsets[1] = $File->get_long($offset); $offset += 4;$counter1 = 0;
			}
		}
		
		# SIS version 2.00 extended fields
		if ($Dump->{Unicode})
		{
		# offset to uncompressed sizes
			for ($counter1 = $Dump->{NumLanguages};$counter1 >=1;$counter1--)
		    {
		    $fileoffsets[$counter1] = $File->get_long($offset); $offset += 4;
		    }
		
		    
		# skip offset to mime type
		# mime type is length-offset pair 
		# length is 0 if no mime type and offset is set to next entry 
		# in string block , this is at the end of the file names
		# 
			$offset += 8;
		    
		}

		$pkglinetype = ($pkglinetype?"LanguageFile":"File");
		
		if ($Dump->{Unicode})
		{      #   0          1          2            3          4          5              6                 7          8
		return ($pkglinetype,$filetype,$filedetails,$Sourcename,$Destname,[@filesizes],[@filecompoffsets],[@fileoffsets],$offset);
			
		}
		else
		{      #   0          1          2            3          4          5              6              7     h
		return ($pkglinetype,$filetype,$filedetails,$Sourcename,$Destname,[@filesizes],[@filecompoffsets],$offset);
		}
    }
    if ($pkglinetype == 2)
    {
	    $pkglinetype = "Option block";
        my ($NumberOfOptions,@OptionsList,$Sourceoptions,$Destoptions,$n);
	    $NumberOfOptions = $File->get_long($offset);$offset += 4;
	    
	    for ($n = 1;$n <= $NumberOfOptions;$n++)
	    {
		$OptionsList[$n] = get_string_got_lenlong($File,$File->get_long($offset+4),$File->get_long($offset));
		$offset+=8;   
	    if ($Dump->{Unicode})
			      {
			      $OptionsList[$n] = Unicode::String->new($OptionsList[$n]);
			      $OptionsList[$n] = latin1($OptionsList[$n]);
			      $OptionsList[$n] =~ s/\c@//g;
		          }
		
		}
		# skip selected option bitflags
		$offset+=16;
	    @OptionsList = reverse @OptionsList;
		pop @OptionsList;
		$Destoptions =  join ",",@OptionsList;
	    $Sourceoptions = "!({" . (join "\"},{\"",@OptionsList) . "\"})";
	    if ($Dump->{Unicode})
		{   #   0          1          2            3             4             5              6                 7          8
		return (0,"Option Block",$filedetails,$Sourceoptions,$Destoptions,[@filesizes],[@filecompoffsets],[@fileoffsets],$offset);
			
		}
		else
		{   #   0         1          2            3               4           5              6              7    
		return (0,"Option Block",$filedetails,$Sourceoptions,$Destoptions,[@filesizes],[@filecompoffsets],$offset);
        } 

    }
    if ($pkglinetype == 3)
    {
	    # fun with if blocks
	    my ($SizeData,$IfString,$temp);
	    $pkglinetype = "If block";# start of IF block
	    # rebuilding a tree
	    ($temp,$offset) = GetIfStuff($File,$offset);
	    $IfString = "If " . $temp;
	    
        #$offset += $SizeData;
	    if ($Dump->{Unicode})
		{      #   0          1          2            3          4             5              6                 7          8
		return (0,"If block",$filedetails,$IfString,$IfString,[@filesizes],[@filecompoffsets],[@fileoffsets],$offset);
			
		}
		else
		{      #   0          1          2            3          4           5              6              7    
		return (0,"If block",$filedetails,$IfString,$IfString,[@filesizes],[@filecompoffsets],$offset);
        } 
	    	    
	}
    if ($pkglinetype == 4)
    {
	    # fun with elseif blocks
	    my ($SizeData,$ElseIfString,$temp);
	    $pkglinetype = "ElseIf block";# start of ELSEIF block
	    ($temp,$offset) = GetIfStuff($File,$offset);
	    $ElseIfString = "ElseIf " . $temp;

	    if ($Dump->{Unicode})
		{      #   0          1          2            3          4             5              6                 7          8
		return (0,"ElseIf block",$filedetails,$ElseIfString,$ElseIfString,[@filesizes],[@filecompoffsets],[@fileoffsets],$offset);
			
		}
		else
		{      #   0          1          2            3          4           5              6              7    
		return (0,"ElseIf block",$filedetails,$ElseIfString,$ElseIfString,[@filesizes],[@filecompoffsets],$offset);
        }
	
	}
    if ($pkglinetype == 5)
    {
	    $pkglinetype = "Else block";# start of ELSE block
	    if ($Dump->{Unicode})
		{      #   0             1               2            3           4             5              6                 7          8
		return (0,"Else block",$filedetails,"Else block","Else block",[@filesizes],[@filecompoffsets],[@fileoffsets],$offset);
			
		}
		else
		{      #   0               1          2                 3         4               5              6              7     h
		return (0,"Else block",$filedetails,"Else block","Else block",[@filesizes],[@filecompoffsets],$offset);
        }
	
	}
  
    if ($pkglinetype == 6)
    {
	
	    $pkglinetype = "EndIf block";# end of IF/ELSEIF/ELSE block
	    if ($Dump->{Unicode})
		{      #   0             1               2            3                    4             5              6                 7          8
		return (0,"End of If block",$filedetails,"End of If block","End of If block",[@filesizes],[@filecompoffsets],[@fileoffsets],$offset);
			
		}
		else
		{      #   0               1          2                 3                4               5              6              7     h
		return (0,"End of If block",$filedetails,"End of If block","End of If block",[@filesizes],[@filecompoffsets],$offset);
        }
	}
    ## Catch all for other line types##
    die " *** Invalid File Line - Corrupt SIS file? ***$ARGV[0]\n"    
}


sub PkgLine2Txt
{
my ($type,$details) = @_;
my $detail; 
		if ($type==0)	{return "Simple File"}        # File to be copied to dest.                 
		if ($type==1)	
		{
			if ($details == 0) {$detail = "Button: Continue"}
			if ($details == 1) {$detail = "Buttons: Yes(Continue)/No(Skip)"}
			if ($details == 2) {$detail = "Buttons: Yes(Continue)/No(Abort)"}
			return ("Text",$detail);
		}        # Text file displayed on install              
		if ($type==2)	
		{
			#component UID
			$detail = sprintf "%#8x",$details;
			return ("Component",$detail);
		}   # Component Sis file                          
		if ($type==3)	
		{
			if ($details == 0) {$detail = "Run on install"}
			if ($details == 1) {$detail = "Run on removal"}
			if ($details == 2) {$detail = "Run on install/removal"}
			return ("Run",$detail)
		}         # File to run, no install                          
		if ($type==4)	{return ("Null/To be created",undef)} # File created by App
		if ($type==5)	{return ("Mime",undef)}        # MIME file                                               
		if ($type==99)	{return ("Not Installed",undef)}# Uninstalled File                 
}


sub GetIfStuff
{
       my ($File,$offset) = @_;
       my ($ifstring,$sizeData,$TotalOffset);
       #first dword is always a condexpr struc
       $sizeData = $File->get_long($offset); $offset += 4;
       $TotalOffset =  $sizeData + $offset;
       
       ($ifstring,$offset) = Conditional2Txt($File,$offset);
       
       if ($offset != $TotalOffset) {die "*** Error parsing if block ***\n"}
       return ($ifstring,$offset);       
}



sub Conditional2Txt
{
        my ($File,$offset) = @_;
        my @binops = (" = "," <> "," > "," < "," >= "," <= "," and "," or ");
      
        my @condVariables = ("Manufacturer","ManufacturerHardwareRev",
        "ManufacturerSoftwareRev","ManufacturerSoftwareBuild","Model",
        "MachineUid","DeviceFamily","DeviceFamilyRev","CPU","CPUArch",
        "CPUABI","CPUSpeed","InvalidConditionVariable","InvalidConditionVariable",
        "SystemTickPeriod","MemoryRAM","MemoryRAMFree","MemoryROM",
        "MemoryPageSize","InvalidConditionVariable","InvalidConditionVariable",
        "PowerBackup","Keyboard","InvalidConditionVariable","InvalidConditionVariable",
        "KeyboardDeviceKeys","KeyboardAppKeys","KeyboardClick",
        "InvalidConditionVariable","InvalidConditionVariable",
        "KeyboardClickVolumeMax","DisplayXPixels","DisplayYPixels","DisplayXTwips",
        "DisplayYTwips","DisplayColors","InvalidConditionVariable","InvalidConditionVariable",        
        "DisplayContrastMax","Backlight","InvalidConditionVariable",
        "Pen","PenX","PenY","PenDisplayOn","PenClick","InvalidConditionVariable",
        "InvalidConditionVariable","PenClickVolumeMax","Mouse","MouseX","MouseY",
        "InvalidConditionVariable","InvalidConditionVariable","InvalidConditionVariable",
        "MouseButtons","InvalidConditionVariable","InvalidConditionVariable","CaseSwitch",
        "InvalidConditionVariable","InvalidConditionVariable","LEDs","InvalidConditionVariable",
        "IntegratedPhone","InvalidConditionVariable","DisplayBrightnessMax",
        "KeyboardBacklightState","AccessoryPower","NumHalAttributes");
	
        $condVariables[0x1000] = "Language";
	$condVariables[0x1001] = "RemoteInstall";
        
        my $conditionstring;
        my $conditional = $File->get_long($offset); $offset += 4;
        
        if ($conditional < 8) #binary op
        {
	       my ($LeftSide,$RightSide) = (""."");
	       $conditionstring = $binops[$conditional];   
           ($LeftSide,$offset) = Conditional2Txt($File,$offset);
           ($RightSide,$offset) = Conditional2Txt($File,$offset);
           return (($LeftSide . $conditionstring . $RightSide), $offset);
        }
        elsif ($conditional < 11)  # Exists() or Devcap()
        {
            my $arg = "";
	        ($conditional == 9)?$conditionstring = " exists(" :$conditionstring = " devcap(" ;
	        ($arg,$offset) = Conditional2Txt($File,$offset);
	        $conditionstring .= "$arg)";
	        return ($conditionstring,$offset);
	    }
        elsif ($conditional == 0xb)  # not op
        {
	       $conditionstring = " not ";
            my $arg = "";
	        ($arg,$offset) = Conditional2Txt($File,$offset);
	        return ($conditionstring,$offset);
        }
        elsif ($conditional == 0xc)  # string
        {
              $conditionstring = get_string_got_lenlong($File,$File->get_long($offset+4),$File->get_long($offset));
	          $offset += 8;
  	          return ($conditionstring,$offset);
        }	        
        elsif ($conditional == 0xd)  # variable
        {
             my $var = $File->get_long($offset);
             # skip over size of union
             $offset += 8;

             if ($var > 0x2000)
             {
	          $var -=0x2000;
	          $conditionstring = "option" . $var;
             }
             else 
             {
	          $conditionstring = $condVariables[$var];
             }
  	          return ($conditionstring,$offset);
        }	        
        elsif ($conditional == 0xe)  # number
        {
             my $num = $File->get_long($offset);
             # skip over size of union
             $offset += 8;
             $conditionstring = sprintf "%2x",$num;
  	          return ($conditionstring,$offset);
        }
  	    else
        {
	        return;
        }
	    
	
}

sub PrintOptions
{
my ($Dump) = @_;
my $o = $Dump->{Options};

   if ($o)
   {

     if ($o & 0x2)  {$Dump->{Distributable} = 1;}
     if ($o & 0x8)  {$Dump->{NoCompress} = 1;}
     if ($o & 0x10) {$Dump->{ShutdownApps} = 1;}
   }
   # print if -e 
   if ($opt_e)
   {
     print "Options-\|\n------- \|\n";
     # unicode determined via uid 2
     if ($Dump->{Unicode})       {print "\t\|\-IsUnicode\n";}      
     if ($Dump->{Distributable}) {print "\t\|\-IsDistributable\n";}
     if ($Dump->{NoCompress})    {print "\t\|\-NoCompress\n";}     
     if ($Dump->{ShutdownApps})  {print "\t\|\-ShutdownApps\n";}       
   } 
 }

sub Type2Txt
{
my ($Dump) = @_;
my $o = $Dump->{Type};

   if ($o)
   {
     if ($o == 0x1)  {$Dump->{Type} = "System";}
     if ($o == 0x2)  {$Dump->{Type} = "Option";}
     if ($o == 0x3)  {$Dump->{Type} = "Config";}
     if ($o == 0x4)  {$Dump->{Type} = "Patch";}
     if ($o == 0x5)  {$Dump->{Type} = "Upgrade";}
   }
   else
   {
   $Dump->{Type} = "App";
   }
}


sub PrintListFiles
{
my ($NumFiles,$SISFiles) = @_;	
my $n;

print "\nFiles\n---------|\n";	
	for ($n = 0; $n <$NumFiles;$n++)
	{
        if ($opt_s || $opt_e)
        {
	printf "   %2d (S)|-%s\n",$n+1,$$SISFiles[$n][3]; # Source filenames option
	printf "      (D)|-%s\n",$$SISFiles[$n][4];	
	    }else
	    {
	printf "   %2d    |-%s\n",$n+1,$$SISFiles[$n][4];	
		}
	    if ($opt_t || $opt_e)
        {
	printf "         |-Type: %s\n",$$SISFiles[$n][1]; # Source filenames option
	    }
	}
}

sub DumpSomeFiles
{
my ($File,$Dump,$FirstFile,$LastFile,$FileDumpArray) = @_;
my ($f,$c,$n,$DumpFile,$CurrentFile,$OutputFile,$OutFH);
my ($vol,$directories,$filename,$dirdepth,@dirs);

     $f = $FirstFile-1;
     print "\n";
for ($c=0;$c <= $LastFile;$c++,$f++)
	{
    ## don't dump conditionals
         unless ($$FileDumpArray[$f][0]) {next}
	#########
    if ($$FileDumpArray[$f][4] ne '') 
	{$DumpFile = "." . substr (($$FileDumpArray[$f][4]),2);}
	else
	{$DumpFile = File::Spec->catfile(@dirs,$$FileDumpArray[$f][3]);}
	$DumpFile =~ tr/\\/\//;
    ### overwrite on request only
	if (-e $DumpFile){unless ($opt_o){next}}
	$CurrentFile = GetFileChunk($File,$$FileDumpArray[$f][6][1],$$FileDumpArray[$f][5][1]);
    $OutputFile = $CurrentFile; # not compressed? - done
	if ($Dump->{Unicode})
       {
	    unless ($Dump->{NoCompress}){$OutputFile = uncompress($CurrentFile)}# decompress stuff
       }
    #dir parsing
    ###############
	($vol,$directories,$filename) = File::Spec->splitpath($DumpFile);
	@dirs = File::Spec->splitdir( $directories );
	$dirdepth = 0;
	for ($n = 1;$n <= @dirs - 2;$n++)
	{
	 unless (-d $dirs[$n]) {mkdir $dirs[$n]}; 
	 chdir $dirs[$n];
	 $dirdepth++;	
	}
	for ($n = $dirdepth;$n>0;$n--)
	{
	 chdir "..";	
	}

    
    open($OutFH,">$DumpFile") or die "Cannot write to $DumpFile \n$!\n";
	binmode $OutFH;
    printf STDERR "dumping to %s.......",File::Spec->catdir(@dirs);
	print $OutFH $OutputFile;
	printf STDERR "%s\n", $filename;
	close $OutFH;
    }	
}


sub GetFileChunk
{
  my ($self, $pos, $Length) = @_;
  $pos -= $self->{start};
  return undef if $pos < 0 || $pos+2 > $self->{len};
  return undef if $pos+$Length > $self->{len};
  return substr($self->{mem}, $pos, $Length);
}

#########################################################
# Friend methods of Disassemble:X86:MemRegion from CPAN
#########################################################

sub get_string_got_lenbyte
{
  my ($self, $pos, $str_len) = @_;
  $pos -= $self->{start};
  my $mem_len = $self->{len};
  return undef if $pos < 0 || $pos >= $mem_len;
  return undef if $pos+$str_len >= $mem_len;
  return substr($self->{mem}, $pos, $str_len);
} # get_string_lenbyte modified to accept length as arg

sub get_string_got_lenword {
  my ($self, $pos, $str_len) = @_;
  $pos -= $self->{start};
  my $mem_len = $self->{len};
  return undef if $pos < 0 || $pos > $mem_len;
  return undef if $pos+$str_len > $mem_len; # length not part of string
  return substr($self->{mem}, $pos, $str_len);
} # get_string_lenword modified to accept length as arg

sub get_string_got_lenlong {
  my ($self, $pos, $str_len) = @_;
  $pos -= $self->{start};
  my $mem_len = $self->{len};
  return undef if $pos < 0 || $pos > $mem_len;
  return undef if $pos+$str_len > $mem_len; # length not part of string
  return substr($self->{mem}, $pos, $str_len);
} # get_string_lenword modified to accept length as arg
